from enum import Enum
from typing import Any

from .key import DependencyKey
from .scope import BaseScope


class FactoryType(Enum):
    GENERATOR = "generator"
    ASYNC_GENERATOR = "async_generator"
    FACTORY = "factory"
    ASYNC_FACTORY = "async_factory"
    VALUE = "value"
    ALIAS = "alias"
    CONTEXT = "context"


class FactoryData:
    __slots__ = ("provides", "scope", "source", "type")

    def __init__(
            self,
            *,
            source: Any,
            provides: DependencyKey,
            scope: BaseScope | None,
            type_: FactoryType,
    ) -> None:
        self.source = source
        self.provides = provides
        self.scope = scope
        self.type = type_



__all__ = (
    "FastStreamProvider",
    "FromDishka",
    "inject",
    "setup_dishka",
)

import warnings
from collections.abc import AsyncIterator, Awaitable, Callable
from contextlib import asynccontextmanager
from typing import (
    Any,
    ParamSpec,
    Protocol,
    TypeAlias,
    TypeVar,
    cast,
)

from faststream import BaseMiddleware, FastStream, context
from faststream.__about__ import (
    __version__ as FASTSTREAM_VERSION,  # noqa: N812
)
from faststream.broker.message import StreamMessage
from faststream.types import DecodedMessage
from faststream.utils.context import ContextRepo

from dishka import AsyncContainer, FromDishka, Provider, Scope, from_context
from dishka.integrations.base import wrap_injection


class FastStreamProvider(Provider):
    context = from_context(ContextRepo, scope=Scope.REQUEST)
    message = from_context(StreamMessage, scope=Scope.REQUEST)


FASTSTREAM_04 = FASTSTREAM_VERSION.startswith("0.4")
FASTSTREAM_05 = FASTSTREAM_VERSION.startswith("0.5")
FASTSTREAM_LATES = not FASTSTREAM_04 and not FASTSTREAM_05

if FASTSTREAM_04:
    warnings.warn(
        "FastStream 0.4 is deprecated and integration will be removed"
        " in the 1.6.0 dishka release.",
        category=DeprecationWarning,
        stacklevel=1,
    )

    from faststream.broker.core.abc import (  # type: ignore[attr-defined]
        BrokerUsecase as BrokerType,
    )

    Application: TypeAlias = FastStream

    try:
        from faststream.broker.fastapi import StreamRouter
    except ImportError:
        pass
    else:
        Application |= StreamRouter  # type: ignore[assignment]

elif FASTSTREAM_05:
    from faststream.broker.core.usecase import BrokerUsecase as BrokerType

    if FASTSTREAM_VERSION < "0.5.16":
        warnings.warn(
            "FastStream < 0.5.16 is deprecated and integration will be removed"
            " in the 1.7.0 dishka release.",
            category=DeprecationWarning,
            stacklevel=1,
        )

        Application: TypeAlias = FastStream  # type: ignore[no-redef,misc]

    else:
        from faststream.asgi import AsgiFastStream

        Application: TypeAlias = FastStream | AsgiFastStream  # type: ignore[no-redef,misc]

    try:
        from faststream.broker.fastapi import StreamRouter
    except ImportError:
        pass
    else:
        Application |= StreamRouter  # type: ignore[assignment]

else:
    from faststream._internal.broker.broker import BrokerUsecase as BrokerType
    from faststream.asgi import AsgiFastStream

    Application: TypeAlias = FastStream | AsgiFastStream  # type: ignore[no-redef,misc]

    try:
        from faststream._internal.fastapi import (  # type: ignore[no-redef]
            StreamRouter,
        )
    except ImportError:
        pass
    else:
        Application |= StreamRouter  # type: ignore[assignment]


class ApplicationLike(Protocol):
    broker: BrokerType[Any, Any]


def setup_dishka(
    container: AsyncContainer,
    app: "Application | ApplicationLike | None" = None,
    broker: "BrokerType[Any, Any] | None" = None,
    *,
    finalize_container: bool = True,
    auto_inject: bool = False,
) -> None:
    """
    Setup dishka integration with FastStream.
    You must provide either app or broker.

    Args:
        container: AsyncContainer instance.
        app: FastStream Application or StreamRouter instance.
        broker: FastStream broker instance.
        finalize_container: bool. Can be used only with app.
        auto_inject: bool.
    """
    if (app and broker) or (not app and not broker):
        raise ValueError(  # noqa: TRY003
            "You must provide either app or broker "
            "to setup dishka integration.",
        )

    if finalize_container:
        if getattr(app, "after_shutdown", None):
            app.after_shutdown(container.close)  # type: ignore[union-attr]

        else:
            warnings.warn(
                "For use `finalize_container=True` "
                "you must provide `app: FastStream | AsgiFastStream` "
                "argument.",
                category=RuntimeWarning,
                stacklevel=2,
            )

    broker = broker or getattr(app, "broker", None)
    assert broker  # noqa: S101

    if FASTSTREAM_04:
        # FastStream 0.4 - 0.5
        broker.middlewares = (
            DishkaMiddleware(container),
            *broker.middlewares,
        )

        if auto_inject:
            warnings.warn(
                """
Auto injection is not supported for FastStream version less than 0.5.0
Please, update your FastStream installation
or use @inject at each subscriber manually.
            """,
                category=RuntimeWarning,
                stacklevel=2,
            )

        return

    elif getattr(broker, "add_middleware", None):
        # FastStream 0.5.6 and higher
        broker.add_middleware(DishkaMiddleware(container))

    else:
        # FastStream 0.5 - 0.5.6
        warnings.warn(
            "FastStream < 0.5.6 is deprecated and integration will be removed"
            " in the 1.7.0 dishka release.",
            category=DeprecationWarning,
            stacklevel=2,
        )

        broker._middlewares = (  # noqa: SLF001
            DishkaMiddleware(container),
            *broker._middlewares,  # noqa: SLF001
        )

        for subscriber in broker._subscribers.values():  # noqa: SLF001
            subscriber._broker_middlewares = (  # noqa: SLF001
                DishkaMiddleware(container),
                *subscriber._broker_middlewares,  # noqa: SLF001
            )

        for publisher in broker._publishers.values():  # noqa: SLF001
            publisher._broker_middlewares = (  # noqa: SLF001
                DishkaMiddleware(container),
                *publisher._broker_middlewares,  # noqa: SLF001
            )

    if auto_inject:
        broker._call_decorators = (  # noqa: SLF001
            inject,
            *broker._call_decorators,  # noqa: SLF001
        )


T = TypeVar("T")
P = ParamSpec("P")


def inject(func: Callable[P, T]) -> Callable[P, T]:
    return wrap_injection(
        func=func,
        is_async=True,
        container_getter=lambda *_: context.get_local("dishka"),
    )


class _DishkaMiddleware(BaseMiddleware):
    def __init__(
        self,
        container: AsyncContainer,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        self.container = container
        super().__init__(*args, **kwargs)

    if FASTSTREAM_04:

        @asynccontextmanager
        async def consume_scope(  # type: ignore[override]
            self,
            *args: Any,
            **kwargs: Any,
        ) -> AsyncIterator[DecodedMessage]:
            async with self.container({
                ContextRepo: context,
            }) as request_container:
                with context.scope("dishka", request_container):
                    async with super().consume_scope(  # type: ignore[attr-defined]
                        *args,
                        **kwargs,
                    ) as result:
                        yield result

    elif FASTSTREAM_05:

        async def consume_scope(  # type: ignore[misc]
            self,
            call_next: Callable[[Any], Awaitable[Any]],
            msg: StreamMessage[Any],
        ) -> AsyncIterator[DecodedMessage]:
            async with self.container(
                {
                    StreamMessage: msg,
                    ContextRepo: context,
                    type(msg): msg,
                },
            ) as request_container:
                with context.scope("dishka", request_container):
                    return cast(
                        AsyncIterator[DecodedMessage],
                        await call_next(msg),
                    )

    else:
        async def consume_scope(  # type: ignore[misc]
            self,
            call_next: Callable[[Any], Awaitable[Any]],
            msg: StreamMessage[Any],
        ) -> AsyncIterator[DecodedMessage]:
            async with self.container(
                {
                    StreamMessage: msg,
                    ContextRepo: context,
                    type(msg): msg,
                },
            ) as request_container:
                with self.context.scope("dishka", request_container):  # type: ignore[attr-defined]
                    return cast(
                        AsyncIterator[DecodedMessage],
                        await call_next(msg),
                    )


class DishkaMiddleware:
    def __init__(self, container: AsyncContainer) -> None:
        self.container = container

    def __call__(self, *args: Any, **kwargs: Any) -> "_DishkaMiddleware":
        return _DishkaMiddleware(self.container, *args, **kwargs)



from abc import abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import Protocol


class GroupType(Enum):
    SCOPE = "SCOPE"
    COMPONENT = "COMPONENT"

class NodeType(Enum):
    CONTEXT = "Context"
    FACTORY = "Factory"
    DECORATOR = "Decorator"
    ALIAS = "Alias"


@dataclass
class Node:
    id: str
    name: str
    dependencies: list[str]
    type: NodeType
    is_protocol: bool
    source_name: str


@dataclass
class Group:
    id: str
    name: str
    children: list["Group"]
    nodes: list[Node]
    type: GroupType


class Renderer(Protocol):
    @abstractmethod
    def render(self, groups: list[Group]) -> str:
        raise NotImplementedError



from __future__ import annotations

import sys
from collections.abc import Sequence
from typing import Any, ForwardRef, TypeVar, get_args, get_origin

from dishka._adaptix.type_tools.basic_utils import eval_forward_ref, is_generic
from dishka.exception_base import DishkaError


def eval_maybe_forward(t: Any, wrapper: Any) -> Any:
    if isinstance(t, str):
        t = ForwardRef(t)
    elif not isinstance(t, ForwardRef):
        return t
    return eval_forward_ref(
        vars(sys.modules[wrapper.__module__]),
        t,
    )


def eval_maybe_forward_many(args: Sequence[Any], wrapper: Any) -> list[type]:
    return [
        eval_maybe_forward(t, wrapper) for t in args
    ]


class _TypeMatcher:
    def __init__(self) -> None:
        self.type_var_subst: dict[TypeVar, Any] = {}

    def _is_broader_or_same_generic(self, t1: Any, t2: Any) -> bool:
        origin1 = eval_maybe_forward(get_origin(t1) or t1, t1)
        origin2 = eval_maybe_forward(get_origin(t2) or t2, t2)
        if origin1 is not origin2:
            return False
        args1 = eval_maybe_forward_many(get_args(t1), t1)
        params1 = t1.__parameters__
        if len(args1) != len(params1):
            args1 += params1[len(args1):]
        args2 = eval_maybe_forward_many(get_args(t2), t2)
        params2 = t1.__parameters__
        if len(args2) != len(params2):
            args2 += params1[len(args2):]

        if len(args1) != len(args2):
            return False
        for a1, a2 in zip(args1, args2, strict=False):
            if not self.is_broader_or_same_type(a1, a2):
                return False
        return True

    def _is_bound_broader(self, t1: TypeVar, t2: Any) -> bool:
        bound1 = eval_maybe_forward(t1.__bound__, t1)
        if get_origin(bound1):
            raise UnsupportedGenericBoundsError(t1)
        if not isinstance(t2, TypeVar):
            if get_origin(t2):
                t2 = get_origin(t2)
            return issubclass(t2, bound1)
        if t2.__bound__:
            bound2 = eval_maybe_forward(t2.__bound__, t2)
            return issubclass(bound2, bound1)
        return False

    def _is_constraint_broader(self, t1: TypeVar, t2: Any) -> bool:
        constraints1 = eval_maybe_forward_many(t1.__constraints__, t1)
        if not isinstance(t2, TypeVar):
            return t2 in constraints1
        if t2.__constraints__:
            constraints2 = eval_maybe_forward_many(t2.__constraints__, t2)
            return set(constraints1) >= set(constraints2)
        return False

    def _is_broader_or_same_type_var(self, t1: TypeVar, t2: Any) -> bool:
        if t1 in self.type_var_subst:
            return self.type_var_subst[t1] == t2  # type: ignore[no-any-return]
        self.type_var_subst[t1] = t2

        if not t1.__bound__ and not t1.__constraints__:
            return True
        if t1.__bound__:
            return self._is_bound_broader(t1, t2)
        elif t1.__constraints__:
            return self._is_constraint_broader(t1, t2)
        return False

    def is_broader_or_same_type(self, t1: Any, t2: Any) -> bool:
        if t1 == t2:
            return True
        if get_origin(t1) is not None or is_generic(t1):
            return self._is_broader_or_same_generic(t1, t2)
        elif isinstance(t1, TypeVar):
            return self._is_broader_or_same_type_var(t1, t2)
        return False


def is_broader_or_same_type(t1: Any, t2: Any) -> bool:
    return _TypeMatcher().is_broader_or_same_type(t1, t2)


def get_typevar_replacement(t1: Any, t2: Any) -> dict[TypeVar, Any]:
    matcher = _TypeMatcher()
    matcher.is_broader_or_same_type(t1, t2)
    return matcher.type_var_subst


class UnsupportedGenericBoundsError(TypeError, DishkaError):
    def __init__(self, bounds: TypeVar) -> None:
        self.bounds = bounds

    def __str__(self) -> str:
        return f"Generic bounds are not supported: {self.bounds}"



from collections.abc import Sequence

from dishka.entities.factory_type import FactoryData
from dishka.exception_base import DishkaError
from dishka.text_rendering import get_name
from dishka.text_rendering.path import PathRenderer
from dishka.text_rendering.suggestion import render_suggestions_for_missing
from .entities.key import DependencyKey
from .entities.scope import BaseScope

try:
    from builtins import (  # type: ignore[attr-defined, unused-ignore]
        ExceptionGroup,
    )

except ImportError:
    from exceptiongroup import (  # type: ignore[no-redef, import-not-found, unused-ignore]
        ExceptionGroup,
    )

_renderer = PathRenderer()


class ExitError(
    ExceptionGroup[Exception],  # type: ignore[misc, unused-ignore]
    DishkaError,
):
    pass


class NoContextValueError(DishkaError):
    pass


class UnsupportedFactoryError(DishkaError):
    pass


class InvalidGraphError(DishkaError):
    pass


class NoFactoryError(DishkaError):
    def __init__(
            self,
            requested: DependencyKey,
            path: Sequence[FactoryData] = (),
            suggest_other_scopes: Sequence[FactoryData] = (),
            suggest_other_components: Sequence[FactoryData] = (),
    ) -> None:
        self.requested = requested
        self.path = list(path)
        self.suggest_other_scopes = suggest_other_scopes
        self.suggest_other_components = suggest_other_components
        self.scope: BaseScope | None = None

    def add_path(self, requested_by: FactoryData) -> None:
        self.path.insert(0, requested_by)

    def __str__(self) -> str:
        suggestion = render_suggestions_for_missing(
            requested_for=self.path[-1] if self.path else None,
            requested_key=self.requested,
            suggest_other_scopes=self.suggest_other_scopes,
            suggest_other_components=self.suggest_other_components,
        )
        if suggestion:
            suggestion = f"Hint:{suggestion}"
        requested_name = get_name(
            self.requested.type_hint, include_module=False,
        )
        if self.path:
            requested_repr = (
                f"({requested_name}, "
                f"component={self.requested.component!r})"
            )
            return (
                f"Cannot find factory for {requested_repr}. "
                f"It is missing or has invalid scope.\n"
            ) + _renderer.render(self.path, self.requested) + suggestion
        else:
            requested_repr = (
                f"({requested_name}, "
                f"component={self.requested.component!r}, "
                f"scope={self.scope!s})"
            )
            return (
                f"Cannot find factory for {requested_repr}. "
                f"Check scopes in your providers. "
                f"It is missing or has invalid scope."
            ) + suggestion


class AliasedFactoryNotFoundError(ValueError, DishkaError):
    def __init__(
            self, dependency: DependencyKey, alias: FactoryData,
    ) -> None:
        self.dependency = dependency
        self.alias_provider = alias.provides

    def __str__(self) -> str:
        return (
            f"Factory for {self.dependency} "
            f"aliased from {self.alias_provider} is not found"
        )


class NoChildScopesError(ValueError, DishkaError):
    def __str__(self) -> str:
        return "No child scopes found"


class NoNonSkippedScopesError(ValueError, DishkaError):
    def __str__(self) -> str:
        return "No non-skipped scopes found."


class ChildScopeNotFoundError(ValueError, DishkaError):
    def __init__(
            self,
            assumed_child_scope: BaseScope | None,
            current_scope: BaseScope | None,
    ) -> None:
        self.child_scope = assumed_child_scope
        self.current_scope = current_scope

    def __str__(self) -> str:
        return (
            f"Cannot find {self.child_scope} as a "
            f"child of current {self.current_scope}"
        )


class UnknownScopeError(InvalidGraphError):
    def __init__(
            self,
            scope: BaseScope | None,
            expected: type[BaseScope],
            extend_message: str = "",
    ) -> None:
        self.scope = scope
        self.expected = expected
        self.extend_message = extend_message

    def __str__(self) -> str:
        return " ".join((
            f"Scope {self.scope} is unknown, "
            f"expected one of {self.expected}",
            self.extend_message,
        ))


class CycleDependenciesError(InvalidGraphError):
    def __init__(self, path: Sequence[FactoryData]) -> None:
        self.path = path

    def __str__(self) -> str:
        if len(self.path) == 1:
            hint = " Did you mean @decorate instead of @provide?"
        else:
            hint = ""
        details = _renderer.render(self.path)
        return f"Cycle dependencies detected.{hint}\n{details}"


class GraphMissingFactoryError(NoFactoryError, InvalidGraphError):
    pass


class ImplicitOverrideDetectedError(InvalidGraphError):
    def __init__(self, new: FactoryData, existing: FactoryData) -> None:
        self.new = new
        self.existing = existing

    def __str__(self) -> str:
        new_name = get_name(self.new.source, include_module=False)
        existing_name = get_name(self.existing.source, include_module=False)
        return (
            f"Detected multiple factories for {self.new.provides} "
            f"while `override` flag is not set.\n"
            "Hint:\n"
            f" * Try specifying `override=True` for {new_name}\n"
            f" * Try removing factory {existing_name} or {new_name}\n"
        )


class NothingOverriddenError(InvalidGraphError):
    def __init__(self, factory: FactoryData) -> None:
        self.factory = factory

    def __str__(self) -> str:
        name = get_name(self.factory.source, include_module=False)
        return (
            f"Overriding factory found for {self.factory.provides}, "
            "but there is nothing to override.\n"
            "Hint:\n"
            f" * Try removing override=True from {name}\n"
            f" * Check the order of providers\n"
        )



__all__ = [
    "Alias",
    "CompositeDependencySource",
    "ContextVariable",
    "Decorator",
    "DependencySource",
    "Factory",
    "context_stub",
    "ensure_composite",
]

from .alias import Alias
from .composite import (
    CompositeDependencySource,
    DependencySource,
    ensure_composite,
)
from .context_var import ContextVariable, context_stub
from .decorator import Decorator
from .factory import Factory



from dishka.exception_base import DishkaError


class InvalidInjectedFuncTypeError(DishkaError):
    def __str__(self) -> str:
        return (
            "An async container cannot be used in a synchronous context."
        )



from collections.abc import AsyncIterator, Iterator
from contextlib import contextmanager
from typing import TypeVar

from dishka import (
    DEFAULT_COMPONENT,
    AsyncContainer,
    Container,
    make_async_container,
    make_container,
)
from dishka.provider import BaseProvider


def _container_provider(container: Container | AsyncContainer) -> BaseProvider:
    container_provider = BaseProvider(component=DEFAULT_COMPONENT)
    container_provider.factories.extend(container.registry.factories.values())
    for registry in container.child_registries:
        container_provider.factories.extend(registry.factories.values())
    return container_provider


CT = TypeVar("CT", Container, AsyncContainer)
def _swap(c1: CT, c2: CT) -> None:
    for attr in type(c1).__slots__:
        tmp = getattr(c1, attr)
        setattr(c1, attr, getattr(c2, attr))
        setattr(c2, attr, tmp)

@contextmanager
def override(container: Container, *providers: BaseProvider) -> Iterator[None]:
    new_container = make_container(_container_provider(container), *providers)
    _swap(container, new_container)
    yield
    container.close()
    _swap(new_container, container)



@contextmanager
async def override_async(
        container: AsyncContainer, *providers: BaseProvider,
) -> AsyncIterator[None]:
    new_container = make_async_container(_container_provider(container), *providers)
    _swap(container, new_container)
    yield
    await container.close()
    _swap(new_container, container)





from __future__ import annotations

import warnings
from collections.abc import Callable, MutableMapping
from contextlib import AbstractContextManager
from threading import Lock
from types import TracebackType
from typing import Any, TypeVar, cast, overload

from dishka.entities.component import DEFAULT_COMPONENT, Component
from dishka.entities.factory_type import FactoryType
from dishka.entities.key import DependencyKey
from dishka.entities.scope import BaseScope, Scope
from .container_objects import Exit
from .context_proxy import ContextProxy
from .dependency_source import Factory
from .entities.validation_settigs import DEFAULT_VALIDATION, ValidationSettings
from .exceptions import (
    ChildScopeNotFoundError,
    ExitError,
    NoChildScopesError,
    NoFactoryError,
    NoNonSkippedScopesError,
)
from .provider import BaseProvider
from .registry import Registry
from .registry_builder import RegistryBuilder

T = TypeVar("T")


class Container:
    __slots__ = (
        "_cache",
        "_context",
        "_exits",
        "child_registries",
        "close_parent",
        "lock",
        "parent_container",
        "registry",
    )

    def __init__(
            self,
            registry: Registry,
            *child_registries: Registry,
            parent_container: Container | None = None,
            context: dict[Any, Any] | None = None,
            lock_factory: Callable[
                [], AbstractContextManager[Any],
            ] | None = None,
            close_parent: bool = False,
    ):
        self.registry = registry
        self.child_registries = child_registries
        self._context = {CONTAINER_KEY: self}
        if context:
            for key, value in context.items():
                if not isinstance(key, DependencyKey):
                    key = DependencyKey(  # noqa: PLW2901
                        key,
                        DEFAULT_COMPONENT,
                    )
                self._context[key] = value
        self._cache = {**self._context}
        self.parent_container = parent_container

        self.lock: AbstractContextManager[Any] | None
        if lock_factory:
            self.lock = lock_factory()
        else:
            self.lock = None
        self._exits: list[Exit] = []
        self.close_parent = close_parent

    @property
    def scope(self) -> BaseScope:
        return self.registry.scope

    @property
    def context(self) -> MutableMapping[DependencyKey, Any]:
        warnings.warn(
            "`container.context` is deprecated",
            DeprecationWarning,
            stacklevel=2,
        )
        return ContextProxy(cache=self._cache, context=self._context)

    def __call__(
            self,
            context: dict[Any, Any] | None = None,
            lock_factory: Callable[
                [], AbstractContextManager[Any],
            ] | None = None,
            scope: BaseScope | None = None,
    ) -> ContextWrapper:
        """
        Prepare container for entering the inner scope.
        :param context: Data which will available in inner scope
        :param lock_factory: Callable to create lock instance or None
        :param scope: target scope or None to enter next non-skipped scope
        :return: context manager for inner scope
        """
        if not self.child_registries:
            raise NoChildScopesError
        child = Container(
            *self.child_registries,
            parent_container=self,
            context=context,
            lock_factory=lock_factory,
        )
        if scope is None:
            while child.registry.scope.skip:
                if not child.child_registries:
                    raise NoNonSkippedScopesError
                child = Container(
                    *child.child_registries,
                    parent_container=child,
                    context=context,
                    lock_factory=lock_factory,
                    close_parent=True,
                )
        else:
            while child.registry.scope is not scope:
                if not child.child_registries:
                    raise ChildScopeNotFoundError(scope, self.registry.scope)
                child = Container(
                    *child.child_registries,
                    parent_container=child,
                    context=context,
                    lock_factory=lock_factory,
                    close_parent=True,
                )
        return ContextWrapper(child)

    @overload
    def get(
            self,
            dependency_type: type[T],
            component: Component | None = DEFAULT_COMPONENT,
    ) -> T:
        ...

    @overload
    def get(
            self,
            dependency_type: Any,
            component: Component | None = DEFAULT_COMPONENT,
    ) -> Any:
        ...

    def get(
            self,
            dependency_type: Any,
            component: Component | None = DEFAULT_COMPONENT,
    ) -> Any:
        lock = self.lock
        key = DependencyKey(dependency_type, component)
        try:
            if not lock:
                return self._get_unlocked(key)
            with lock:
                return self._get_unlocked(key)
        except NoFactoryError as e:
            e.scope = self.scope
            raise

    def _get(self, key: DependencyKey) -> Any:
        lock = self.lock
        if not lock:
            return self._get_unlocked(key)
        with lock:
            return self._get_unlocked(key)

    def _get_unlocked(self, key: DependencyKey) -> Any:
        if key in self._cache:
            return self._cache[key]
        compiled = self.registry.get_compiled(key)
        if not compiled:
            if not self.parent_container:
                raise NoFactoryError(key)
            return self.parent_container._get(key)  # noqa: SLF001
        try:
            return compiled(self._get_unlocked, self._exits, self._cache)
        except NoFactoryError as e:
            # cast is needed because registry.get_factory will always
            # return Factory. This happens because registry.get_compiled
            # uses the same method and returns None if the factory is not found
            # If None is returned, then go to the parent container
            e.add_path(cast(Factory, self.registry.get_factory(key)))
            raise

    def close(self, exception: BaseException | None = None) -> None:
        errors = []
        for exit_generator in self._exits[::-1]:
            try:
                if exit_generator.type is FactoryType.GENERATOR:
                    exit_generator.callable.send(exception)  # type: ignore[attr-defined]
            except StopIteration:  # noqa: PERF203
                pass
            except Exception as err:  # noqa: BLE001
                errors.append(err)
        self._cache = {**self._context}
        if self.close_parent and self.parent_container:
            try:
                self.parent_container.close(exception)
            except Exception as err:  # noqa: BLE001
                errors.append(err)

        if errors:
            raise ExitError("Cleanup context errors", errors)  # noqa: TRY003


class ContextWrapper:
    __slots__ = ("container",)

    def __init__(self, container: Container):
        self.container = container

    def __enter__(self) -> Container:
        return self.container

    def __exit__(
        self,
        exc_type: type[BaseException] | None = None,
        exc_val: BaseException | None = None,
        exc_tb: TracebackType | None = None,
    ) -> None:
        self.container.close(exception=exc_val)


def make_container(
        *providers: BaseProvider,
        scopes: type[BaseScope] = Scope,
        context: dict[Any, Any] | None = None,
        lock_factory: Callable[[], AbstractContextManager[Any]] | None = Lock,
        skip_validation: bool = False,
        start_scope: BaseScope | None = None,
        validation_settings: ValidationSettings = DEFAULT_VALIDATION,
) -> Container:
    registries = RegistryBuilder(
        scopes=scopes,
        container_key=CONTAINER_KEY,
        providers=providers,
        skip_validation=skip_validation,
        validation_settings=validation_settings,
    ).build()
    container = Container(
        *registries,
        context=context,
        lock_factory=lock_factory,
    )
    if start_scope is None:
        while container.registry.scope.skip:
            container = Container(
                *container.child_registries,
                parent_container=container,
                context=context,
                lock_factory=lock_factory,
                close_parent=True,
            )
    else:
        while container.registry.scope is not start_scope:
            container = Container(
                *container.child_registries,
                parent_container=container,
                context=context,
                lock_factory=lock_factory,
                close_parent=True,
            )
    return container


CONTAINER_KEY = DependencyKey(Container, DEFAULT_COMPONENT)



from __future__ import annotations

import sys
import threading
from typing import TYPE_CHECKING, Any, Generic, TypeVar

__all__ = ["AnyOf", "ProvideMultiple"]


if sys.version_info >= (3, 11):
    from typing import TypeVarTuple, Unpack

    Variants = TypeVarTuple("Variants")

    class ProvideMultiple(Generic[Unpack[Variants]]):
        pass
else:
    # in this case we simulate generics with variadic typevars
    # by suppressing builtin checks of parameters length
    Variants = TypeVar("Variants")
    provides_lock = threading.Lock()

    class ProvideMultiple(Generic[Variants]):
        def __class_getitem__(cls, item: tuple[Any]) -> Any:
            with provides_lock:
                cls.__parameters__ = [Variants]*len(item)  # type: ignore[attr-defined, misc]
                return super().__class_getitem__(item)  # type: ignore[misc]


if TYPE_CHECKING:
    from typing import Union as AnyOf
else:
    AnyOf = ProvideMultiple



from collections.abc import Callable
from typing import Any, TypeVar, get_args, get_origin

from ._adaptix.type_tools.fundamentals import get_type_vars
from .container_objects import CompiledFactory
from .dependency_source import (
    Factory,
)
from .dependency_source.type_match import (
    get_typevar_replacement,
    is_broader_or_same_type,
)
from .entities.factory_type import FactoryType
from .entities.key import DependencyKey
from .entities.scope import BaseScope
from .factory_compiler import compile_factory


class Registry:
    __slots__ = (
        "compiled",
        "compiled_async",
        "factories",
        "has_fallback",
        "scope",
    )

    def __init__(self, scope: BaseScope, *, has_fallback: bool) -> None:
        self.scope = scope
        self.factories: dict[DependencyKey, Factory] = {}
        self.compiled: dict[DependencyKey, Callable[..., Any]] = {}
        self.compiled_async: dict[DependencyKey, Callable[..., Any]] = {}
        self.has_fallback = has_fallback

    def add_factory(
            self,
            factory: Factory,
            provides: DependencyKey | None = None,
    ) -> None:
        if provides is None:
            provides = factory.provides
        self.factories[provides] = factory

    def get_compiled(
            self, dependency: DependencyKey,
    ) -> CompiledFactory | None:
        try:
            return self.compiled[dependency]
        except KeyError:
            factory = self.get_factory(dependency)
            if not factory:
                return None
            compiled = compile_factory(factory=factory, is_async=False)
            self.compiled[dependency] = compiled
            return compiled

    def get_compiled_async(
            self, dependency: DependencyKey,
    ) -> CompiledFactory | None:
        try:
            return self.compiled[dependency]
        except KeyError:
            factory = self.get_factory(dependency)
            if not factory:
                return None
            compiled = compile_factory(factory=factory, is_async=True)
            self.compiled[dependency] = compiled
            return compiled

    def get_factory(self, dependency: DependencyKey) -> Factory | None:
        try:
            return self.factories[dependency]
        except KeyError:
            origin = get_origin(dependency.type_hint)
            if not origin:
                return None
            if origin is type and self.has_fallback:
                return self._get_type_var_factory(dependency)

            origin_key = DependencyKey(origin, dependency.component)
            factory = self.factories.get(origin_key)
            if not factory:
                return None
            if not is_broader_or_same_type(
                    factory.provides.type_hint,
                    dependency.type_hint,
            ):
                return None
            factory = self._specialize_generic(factory, dependency)
            self.factories[dependency] = factory
            return factory

    def _get_type_var_factory(self, dependency: DependencyKey) -> Factory:
        args = get_args(dependency.type_hint)
        if args:
            typevar = args[0]
        else:
            typevar = Any
        return Factory(
            scope=self.scope,
            dependencies=[],
            kw_dependencies={},
            provides=DependencyKey(type[typevar], dependency.component),
            type_=FactoryType.FACTORY,
            is_to_bind=False,
            cache=False,
            override=False,
            source=lambda: typevar,
        )

    def _specialize_generic(
            self, factory: Factory, dependency_key: DependencyKey,
    ) -> Factory:
        params_replacement = get_typevar_replacement(
            factory.provides.type_hint,
            dependency_key.type_hint,
        )
        new_dependencies: list[DependencyKey] = []
        for source_dependency in factory.dependencies:
            hint = source_dependency.type_hint
            if isinstance(hint, TypeVar):
                hint = params_replacement[hint]
            elif get_origin(hint) and (type_vars := get_type_vars(hint)):
                hint = hint[tuple(
                    params_replacement[param]
                    for param in type_vars
                )]
            new_dependencies.append(DependencyKey(
                hint, source_dependency.component,
            ))
        new_kw_dependencies: dict[str, DependencyKey] = {}
        for name, source_dependency in factory.kw_dependencies.items():
            hint = source_dependency.type_hint
            if isinstance(hint, TypeVar):
                hint = params_replacement[hint]
            elif get_origin(hint) and (type_vars := get_type_vars(hint)):
                hint = hint[tuple(
                    params_replacement[param]
                    for param in type_vars
                )]
            new_kw_dependencies[name] = DependencyKey(
                hint, source_dependency.component,
            )
        return Factory(
            source=factory.source,
            provides=dependency_key,
            dependencies=new_dependencies,
            kw_dependencies=new_kw_dependencies,
            is_to_bind=factory.is_to_bind,
            type_=factory.type,
            scope=factory.scope,
            cache=factory.cache,
            override=factory.override,
        )



from collections.abc import Sequence

from dishka.entities.factory_type import FactoryData
from dishka.entities.key import DependencyKey
from .name import get_name


def render_suggestions_for_missing(
        requested_for: FactoryData | None,
        requested_key: DependencyKey,
        suggest_other_scopes: Sequence[FactoryData],
        suggest_other_components: Sequence[FactoryData],
) -> str:
    suggestion = ""
    if suggest_other_scopes:
        scopes = " or ".join(
            str(factory.scope)
            for factory in suggest_other_scopes
        )
        if requested_for:
            srcname = get_name(requested_for.source, include_module=False)
            suggestion += f"\n * Try changing scope of `{srcname}` to {scopes}"
        else:
            suggestion += f"\n * Check if you forgot enter {scopes}"

    if suggest_other_components:
        dep_name = get_name(requested_key.type_hint, include_module=True)
        components_str = " or ".join(
            f"Annotated[{dep_name}, FromComponent({f.provides.component!r})]"
            for f in suggest_other_components
        )
        suggestion += f"\n * Try using {components_str}"
    return suggestion



__all__ = [
    "DishkaTask",
    "FromDishka",
    "inject",
    "setup_dishka",
]

from collections.abc import Callable
from typing import Final, ParamSpec, TypeVar

from celery import Celery, Task, current_app
from celery.utils.functional import head_from_fun

from dishka import Container, FromDishka
from dishka.integrations.base import is_dishka_injected, wrap_injection

CONTAINER_NAME: Final = "dishka_container"


T = TypeVar("T")
P = ParamSpec("P")


def inject(func: Callable[P, T]) -> Callable[P, T]:
    return wrap_injection(
        func=func,
        is_async=False,
        container_getter=lambda args, kwargs: current_app.conf[
            CONTAINER_NAME
        ],
        manage_scope=True,
    )


def setup_dishka(container: Container, app: Celery):
    app.conf[CONTAINER_NAME] = container


class DishkaTask(Task):
    def __init__(self) -> None:
        super().__init__()

        run = self.run

        if not is_dishka_injected(run):
            injected_func = inject(run)
            self.run = injected_func

            self.__header__ = head_from_fun(injected_func)



from collections.abc import (
    AsyncIterator,
    Awaitable,
    Callable,
    Iterator,
    Sequence,
)
from dataclasses import dataclass
from enum import Enum
from inspect import (
    Parameter,
    Signature,
    _ParameterKind,
    isasyncgenfunction,
    iscoroutinefunction,
    isgeneratorfunction,
    signature,
)
from typing import (
    Annotated,
    Any,
    Literal,
    ParamSpec,
    TypeAlias,
    TypeVar,
    cast,
    get_args,
    get_origin,
    get_type_hints,
    overload,
)

from dishka.async_container import AsyncContainer
from dishka.container import Container
from dishka.entities.component import DEFAULT_COMPONENT
from dishka.entities.depends_marker import FromDishka
from dishka.entities.key import DependencyKey, _FromComponent
from dishka.integrations.exceptions import InvalidInjectedFuncTypeError

T = TypeVar("T")
P = ParamSpec("P")
DependencyParser: TypeAlias = Callable[[Parameter, Any], DependencyKey | None]
ContainerGetter: TypeAlias = Callable[[tuple[Any, ...], dict[str, Any]], T]
DependsClass: TypeAlias = cast(
    type | Sequence[type],
    FromDishka | _FromComponent,
)


def _get_auto_injected_async_gen_scoped(
    container_getter: ContainerGetter[AsyncContainer],
    additional_params: Sequence[Parameter],
    dependencies: dict[str, DependencyKey],
    func: Callable[P, AsyncIterator[T]],
) -> Callable[P, AsyncIterator[T]]:
    async def auto_injected_generator(
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> AsyncIterator[T]:
        for param in additional_params:
            kwargs.pop(param.name)

        container = container_getter(args, kwargs)
        async with container() as container:
            solved = {
                name: await container.get(
                    dep.type_hint,
                    component=dep.component,
                )
                for name, dep in dependencies.items()
            }
            async for message in func(*args, **kwargs, **solved):
                yield message

    return auto_injected_generator


def _get_auto_injected_async_gen(
    container_getter: ContainerGetter[AsyncContainer],
    additional_params: Sequence[Parameter],
    dependencies: dict[str, DependencyKey],
    func: Callable[P, AsyncIterator[T]],
) -> Callable[P, AsyncIterator[T]]:
    async def auto_injected_generator(
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> AsyncIterator[T]:
        for param in additional_params:
            kwargs.pop(param.name)

        container = container_getter(args, kwargs)
        solved = {
            name: await container.get(
                dep.type_hint,
                component=dep.component,
            )
            for name, dep in dependencies.items()
        }
        async for message in func(*args, **kwargs, **solved):
            yield message

    return auto_injected_generator


def _get_auto_injected_async_func_scoped(
    container_getter: ContainerGetter[AsyncContainer],
    additional_params: Sequence[Parameter],
    dependencies: dict[str, DependencyKey],
    func: Callable[P, Awaitable[T]],
) -> Callable[P, Awaitable[T]]:
    async def auto_injected_func(*args: P.args, **kwargs: P.kwargs) -> T:
        container = container_getter(args, kwargs)
        for param in additional_params:
            kwargs.pop(param.name)
        async with container() as container:
            solved = {
                name: await container.get(
                    dep.type_hint,
                    component=dep.component,
                )
                for name, dep in dependencies.items()
            }
            return await func(*args, **kwargs, **solved)

    return auto_injected_func


def _get_auto_injected_async_func(
    container_getter: ContainerGetter[AsyncContainer],
    additional_params: Sequence[Parameter],
    dependencies: dict[str, DependencyKey],
    func: Callable[P, Awaitable[T]],
) -> Callable[P, Awaitable[T]]:
    async def auto_injected_func(*args: P.args, **kwargs: P.kwargs) -> T:
        container = container_getter(args, kwargs)
        for param in additional_params:
            kwargs.pop(param.name)
        solved = {
            name: await container.get(
                dep.type_hint,
                component=dep.component,
            )
            for name, dep in dependencies.items()
        }
        return await func(*args, **kwargs, **solved)

    return auto_injected_func


def _get_auto_injected_sync_gen_scoped(
    container_getter: ContainerGetter[Container],
    additional_params: Sequence[Parameter],
    dependencies: dict[str, DependencyKey],
    func: Callable[P, Iterator[T]],
) -> Callable[P, Iterator[T]]:
    def auto_injected_generator(
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Iterator[T]:
        for param in additional_params:
            kwargs.pop(param.name)

        container = container_getter(args, kwargs)
        with container() as container:
            solved = {
                name: container.get(dep.type_hint, component=dep.component)
                for name, dep in dependencies.items()
            }
            yield from func(*args, **kwargs, **solved)

    return auto_injected_generator


def _get_auto_injected_sync_gen(
    container_getter: ContainerGetter[Container],
    additional_params: Sequence[Parameter],
    dependencies: dict[str, DependencyKey],
    func: Callable[P, Iterator[T]],
) -> Callable[P, Iterator[T]]:
    def auto_injected_generator(
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Iterator[T]:
        container = container_getter(args, kwargs)
        for param in additional_params:
            kwargs.pop(param.name)
        solved = {
            name: container.get(dep.type_hint, component=dep.component)
            for name, dep in dependencies.items()
        }
        yield from func(*args, **kwargs, **solved)

    return auto_injected_generator


def _get_auto_injected_sync_func_scoped(
    container_getter: ContainerGetter[Container],
    additional_params: Sequence[Parameter],
    dependencies: dict[str, DependencyKey],
    func: Callable[P, T],
) -> Callable[P, T]:
    def auto_injected_func(*args: P.args, **kwargs: P.kwargs) -> T:
        for param in additional_params:
            kwargs.pop(param.name)

        container = container_getter(args, kwargs)
        with container() as container:
            solved = {
                name: container.get(dep.type_hint, component=dep.component)
                for name, dep in dependencies.items()
            }
            return func(*args, **kwargs, **solved)

    return auto_injected_func


def _get_auto_injected_sync_func(
    container_getter: ContainerGetter[Container],
    additional_params: Sequence[Parameter],
    dependencies: dict[str, DependencyKey],
    func: Callable[P, T],
) -> Callable[P, T]:
    def auto_injected_func(*args: P.args, **kwargs: P.kwargs) -> T:
        container = container_getter(args, kwargs)
        for param in additional_params:
            kwargs.pop(param.name)

        solved = {
            name: container.get(dep.type_hint, component=dep.component)
            for name, dep in dependencies.items()
        }
        return func(*args, **kwargs, **solved)

    return auto_injected_func


def _get_auto_injected_sync_container_async_gen_scoped(
    container_getter: ContainerGetter[Container],
    additional_params: Sequence[Parameter],
    dependencies: dict[str, DependencyKey],
    func: Callable[P, Iterator[T]],
) -> Callable[P, AsyncIterator[T]]:
    async def auto_injected_generator(
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> AsyncIterator[T]:
        for param in additional_params:
            kwargs.pop(param.name)

        container = container_getter(args, kwargs)
        with container() as container:
            solved = {
                name: container.get(dep.type_hint, component=dep.component)
                for name, dep in dependencies.items()
            }
            async for message in func(*args, **kwargs, **solved):
                yield message

    return auto_injected_generator


def _get_auto_injected_sync_container_async_gen(
    container_getter: ContainerGetter[Container],
    additional_params: Sequence[Parameter],
    dependencies: dict[str, DependencyKey],
    func: Callable[P, Iterator[T]],
) -> Callable[P, AsyncIterator[T]]:
    async def auto_injected_generator(
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> AsyncIterator[T]:
        container = container_getter(args, kwargs)
        for param in additional_params:
            kwargs.pop(param.name)
        solved = {
            name: container.get(dep.type_hint, component=dep.component)
            for name, dep in dependencies.items()
        }
        async for message in func(*args, **kwargs, **solved):
            yield message

    return auto_injected_generator


def _get_auto_injected_sync_container_async_func_scoped(
    container_getter: ContainerGetter[Container],
    additional_params: Sequence[Parameter],
    dependencies: dict[str, DependencyKey],
    func: Callable[P, T],
) -> Callable[P, T]:
    async def auto_injected_func(*args: P.args, **kwargs: P.kwargs) -> T:
        for param in additional_params:
            kwargs.pop(param.name)

        container = container_getter(args, kwargs)
        with container() as container:
            solved = {
                name: container.get(dep.type_hint, component=dep.component)
                for name, dep in dependencies.items()
            }
            return await func(*args, **kwargs, **solved)

    return auto_injected_func


def _get_auto_injected_sync_container_async_func(
    container_getter: ContainerGetter[Container],
    additional_params: Sequence[Parameter],
    dependencies: dict[str, DependencyKey],
    func: Callable[P, T],
) -> Callable[P, T]:
    async def auto_injected_func(*args: P.args, **kwargs: P.kwargs) -> T:
        container = container_getter(args, kwargs)
        for param in additional_params:
            kwargs.pop(param.name)

        solved = {
            name: container.get(dep.type_hint, component=dep.component)
            for name, dep in dependencies.items()
        }
        return await func(*args, **kwargs, **solved)

    return auto_injected_func


class FunctionType(Enum):
    GENERATOR = "generator"
    ASYNC_GENERATOR = "async_generator"
    CALLABLE = "callable"
    ASYNC_CALLABLE = "async_callable"


@dataclass(frozen=True, slots=True, kw_only=True)
class InjectedFuncType:
    is_async_container: bool
    manage_scope: bool
    func_type: FunctionType

    def __post_init__(self) -> None:
        if self.is_async_container and self.func_type in (
            FunctionType.GENERATOR,
            FunctionType.CALLABLE,
        ):
            raise InvalidInjectedFuncTypeError


_GET_AUTO_INJECTED_FUNC_DICT = {
    InjectedFuncType(
        is_async_container=True,
        manage_scope=True,
        func_type=FunctionType.ASYNC_GENERATOR,
    ): _get_auto_injected_async_gen_scoped,
    InjectedFuncType(
        is_async_container=True,
        manage_scope=False,
        func_type=FunctionType.ASYNC_GENERATOR,
    ): _get_auto_injected_async_gen,
    InjectedFuncType(
        is_async_container=True,
        func_type=FunctionType.ASYNC_CALLABLE,
        manage_scope=True,
    ): _get_auto_injected_async_func_scoped,
    InjectedFuncType(
        is_async_container=True,
        func_type=FunctionType.ASYNC_CALLABLE,
        manage_scope=False,
    ): _get_auto_injected_async_func,
    InjectedFuncType(
        is_async_container=False,
        manage_scope=True,
        func_type=FunctionType.GENERATOR,
    ): _get_auto_injected_sync_gen_scoped,
    InjectedFuncType(
        is_async_container=False,
        manage_scope=False,
        func_type=FunctionType.GENERATOR,
    ): _get_auto_injected_sync_gen,
    InjectedFuncType(
        is_async_container=False,
        manage_scope=True,
        func_type=FunctionType.CALLABLE,
    ): _get_auto_injected_sync_func_scoped,
    InjectedFuncType(
        is_async_container=False,
        manage_scope=False,
        func_type=FunctionType.CALLABLE,
    ): _get_auto_injected_sync_func,
    InjectedFuncType(
        is_async_container=False,
        manage_scope=True,
        func_type=FunctionType.ASYNC_CALLABLE,
    ): _get_auto_injected_sync_container_async_func_scoped,
    InjectedFuncType(
        is_async_container=False,
        manage_scope=False,
        func_type=FunctionType.ASYNC_CALLABLE,
    ): _get_auto_injected_sync_container_async_func,
    InjectedFuncType(
        is_async_container=False,
        manage_scope=True,
        func_type=FunctionType.ASYNC_GENERATOR,
    ): _get_auto_injected_sync_container_async_gen_scoped,
    InjectedFuncType(
        is_async_container=False,
        manage_scope=False,
        func_type=FunctionType.ASYNC_GENERATOR,
    ): _get_auto_injected_sync_container_async_gen,
}


def get_func_type(func: Callable) -> FunctionType:
    if isasyncgenfunction(func):
        return FunctionType.ASYNC_GENERATOR
    elif isgeneratorfunction(func):
        return FunctionType.GENERATOR
    elif iscoroutinefunction(func):
        return FunctionType.ASYNC_CALLABLE
    else:
        return FunctionType.CALLABLE


def default_parse_dependency(
    parameter: Parameter,
    hint: Any,
) -> DependencyKey | None:
    """Resolve dependency type or return None if it is not a dependency."""
    if get_origin(hint) is not Annotated:
        return None
    args = get_args(hint)
    dep = next(
        (arg for arg in args if isinstance(arg, DependsClass)),
        None,
    )
    if not dep:
        return None
    if isinstance(dep, FromDishka | _FromComponent):  # type: ignore[arg-type]
        return DependencyKey(args[0], dep.component)
    else:
        return DependencyKey(args[0], DEFAULT_COMPONENT)


@overload
def wrap_injection(
    *,
    func: Callable[P, T],
    container_getter: ContainerGetter[Container],
    is_async: Literal[False] = False,
    remove_depends: bool = True,
    additional_params: Sequence[Parameter] = (),
    parse_dependency: DependencyParser = default_parse_dependency,
    manage_scope: bool = False,
) -> Callable[P, T]: ...


@overload
def wrap_injection(
    *,
    func: Callable[P, T],
    container_getter: ContainerGetter[AsyncContainer],
    is_async: Literal[True] = True,
    remove_depends: bool = True,
    additional_params: Sequence[Parameter] = (),
    parse_dependency: DependencyParser = default_parse_dependency,
    manage_scope: bool = False,
) -> Callable[P, T]: ...


def wrap_injection(
    *,
    func: Callable[P, T],
    container_getter: ContainerGetter[Container | AsyncContainer],
    is_async: bool = False,
    remove_depends: bool = True,
    additional_params: Sequence[Parameter] = (),
    parse_dependency: DependencyParser = default_parse_dependency,
    manage_scope: bool = False,
) -> Callable[P, T]:
    hints = get_type_hints(func, include_extras=True)
    func_signature = signature(func)

    dependencies = {}
    for name, param in func_signature.parameters.items():
        hint = hints.get(name, Any)
        dep = parse_dependency(param, hint)
        if dep is None:
            continue
        dependencies[name] = dep

    if not dependencies:
        try:
            func.__dishka_injected__ = True  # type: ignore[attr-defined]
        except AttributeError:
            pass
        else:
            return func

    if remove_depends:
        new_annotations = {
            name: hint
            for name, hint in hints.items()
            if name not in dependencies
        }
        new_params = [
            param
            for name, param in func_signature.parameters.items()
            if name not in dependencies
        ]
    else:
        new_annotations = hints.copy()
        new_params = list(func_signature.parameters.values())
    # fix hints
    new_params = [
        param.replace(annotation=hints.get(param.name, param.annotation))
        for param in new_params
    ]

    auto_injected_func: Callable[P, T | Awaitable[T]]
    if additional_params:
        new_params = _add_params(new_params, additional_params)
        for param in additional_params:
            new_annotations[param.name] = param.annotation

    if is_async:
        func = cast(Callable[P, Awaitable[T]], func)
        container_getter = cast(
            ContainerGetter[AsyncContainer],
            container_getter,
        )
    else:
        container_getter = cast(
            ContainerGetter[Container],
            container_getter,
        )

    injected_func_type = InjectedFuncType(
        is_async_container=is_async,
        manage_scope=manage_scope,
        func_type=get_func_type(func),
    )
    get_auto_injected_func = _GET_AUTO_INJECTED_FUNC_DICT[injected_func_type]

    auto_injected_func = get_auto_injected_func(
        func=func,
        dependencies=dependencies,
        additional_params=additional_params,
        container_getter=container_getter,
    )

    auto_injected_func.__dishka_orig_func__ = func
    auto_injected_func.__dishka_injected__ = True  # type: ignore[attr-defined]
    auto_injected_func.__name__ = func.__name__
    auto_injected_func.__qualname__ = func.__qualname__
    auto_injected_func.__doc__ = func.__doc__
    auto_injected_func.__module__ = func.__module__
    auto_injected_func.__annotations__ = new_annotations
    auto_injected_func.__signature__ = Signature(  # type: ignore[attr-defined]
        parameters=new_params,
        return_annotation=func_signature.return_annotation,
    )
    return cast(Callable[P, T], auto_injected_func)


def is_dishka_injected(func: Callable[..., Any]) -> bool:
    return hasattr(func, "__dishka_injected__")


def _add_params(
    params: Sequence[Parameter],
    additional_params: Sequence[Parameter],
):
    params_kind_dict: dict[_ParameterKind, list[Parameter]] = {}

    for param in params:
        params_kind_dict.setdefault(param.kind, []).append(param)

    for param in additional_params:
        params_kind_dict.setdefault(param.kind, []).append(param)

    var_positional = params_kind_dict.get(Parameter.VAR_POSITIONAL, [])
    if len(var_positional) > 1:
        param_names = (param.name for param in var_positional)
        var_positional_names = ", *".join(param_names)
        base_msg = "more than one variadic positional parameter: *"
        msg = base_msg + var_positional_names
        raise ValueError(msg)

    var_keyword = params_kind_dict.get(Parameter.VAR_KEYWORD, [])
    if len(var_keyword) > 1:
        var_keyword_names = ", **".join(param.name for param in var_keyword)
        msg = "more than one variadic keyword parameter: " + var_keyword_names
        raise ValueError(msg)

    positional_only = params_kind_dict.get(Parameter.POSITIONAL_ONLY, [])
    positional_or_keyword = params_kind_dict.get(
        Parameter.POSITIONAL_OR_KEYWORD,
        [],
    )
    keyword_only = params_kind_dict.get(Parameter.KEYWORD_ONLY, [])

    result_params = []
    result_params.extend(positional_only)
    result_params.extend(positional_or_keyword)
    result_params.extend(var_positional)
    result_params.extend(keyword_only)
    result_params.extend(var_keyword)

    return result_params



from typing import Any

from dishka import AsyncContainer, BaseScope, Container, DependencyKey
from dishka._adaptix.type_tools.basic_utils import is_protocol
from dishka.dependency_source import Factory
from dishka.entities.factory_type import FactoryType
from dishka.registry import Registry
from dishka.registry_builder import DECORATED_COMPONENT_PREFIX
from dishka.text_rendering import get_name
from .model import Group, GroupType, Node, NodeType


class Transformer:
    def __init__(self) -> None:
        self.nodes: dict[tuple[DependencyKey, BaseScope], Node] = {}
        self.groups: dict[Any, Group] = {}
        self._counter = 0

    def count(self, prefix: str) -> str:
        self._counter += 1
        return f"{prefix}{self._counter}"

    def _is_empty(self, registry: Registry) -> bool:
        for factory in registry.factories.values():
            if factory.provides.type_hint not in (Container, AsyncContainer):
                return False
        return True

    def _is_decorated(self, key: DependencyKey) -> bool:
        if not key.component:
            return False
        return key.component.startswith(DECORATED_COMPONENT_PREFIX)

    def _node_type(self, factory: Factory) -> NodeType:
        for dep in factory.dependencies:
            if self._is_decorated(dep):
                return NodeType.DECORATOR
        for dep in factory.kw_dependencies.values():
            if self._is_decorated(dep):
                return NodeType.DECORATOR

        if factory.type is FactoryType.ALIAS:
            return NodeType.ALIAS
        elif factory.type is FactoryType.CONTEXT:
            return NodeType.CONTEXT
        else:
            return NodeType.FACTORY

    def _is_decorated_component(self, group: Group) -> bool:
        if group.type is not GroupType.COMPONENT:
            return False
        return group.name.startswith(DECORATED_COMPONENT_PREFIX)

    def _trace_decorator(self, node: Node, target_group: Group) -> None:
        if node.type is not NodeType.DECORATOR:
            return
        for group in self.groups.values():
            if not self._is_decorated_component(group):
                continue
            nodes_to_move = [
                n
                for n in group.nodes
                if n.id in node.dependencies
            ]
            for moved in nodes_to_move:
                group.nodes.remove(moved)
                target_group.nodes.append(moved)
                self._trace_decorator(moved, target_group)

    def _make_factories(
            self, scope: BaseScope, group: Group, registry: Registry,
    ) -> None:
        for key, factory in registry.factories.items():
            group_key = (scope, key.component)
            if group_key in self.groups:
                component_group = self.groups[group_key]
            else:
                component_group = self.groups[group_key] = Group(
                    id=self.count("component"),
                    name=str(key.component),
                    children=[],
                    nodes=[],
                    type=GroupType.COMPONENT,
                )
                group.children.append(component_group)
            node_name = get_name(key.type_hint, include_module=False)
            if factory.type in (FactoryType.CONTEXT, FactoryType.ALIAS):
                source_name = ""
            else:
                source_name = get_name(factory.source, include_module=False)

            node = Node(
                id=self.count("factory"),
                name=node_name,
                dependencies=[],
                type=self._node_type(factory),
                is_protocol=is_protocol(factory.provides.type_hint),
                source_name=source_name,
            )
            self.nodes[key, scope] = node
            component_group.nodes.append(node)

    def _fill_dependencies(
            self, registry: Registry, parent_registries: list[Registry],
    ) -> None:
        # parent registries are passed from current scope to outer one
        for key, factory in registry.factories.items():
            node = self.nodes[key, registry.scope]
            all_deps = (
                    list(factory.dependencies)
                    + list(factory.kw_dependencies.values())
            )
            for dep in all_deps:
                for dep_registry in parent_registries:
                    if dep in dep_registry.factories:
                        break
                else:
                    continue
                dep_node = self.nodes[dep, dep_registry.scope]
                node.dependencies.append(dep_node.id)

    def clean_groups(self, groups: list[Group]) -> list[Group]:
        res = []
        for group in groups:
            group.children = self.clean_groups(group.children)
            if group.children or group.nodes:
                res.append(group)
        return res

    def transform(self, container: Container | AsyncContainer) -> list[Group]:
        registries = [container.registry, *container.child_registries]
        result = []
        for registry in registries:
            if self._is_empty(registry):
                continue
            scope = registry.scope
            group = self.groups[scope] = Group(
                id=self.count("scope"),
                name=str(scope),
                children=[],
                nodes=[],
                type=GroupType.SCOPE,
            )
            result.append(group)
            self._make_factories(scope, group, registry)

        for n, registry in enumerate(registries):
            if self._is_empty(registry):
                continue
            self._fill_dependencies(registry, registries[n::-1])

        for group in self.groups.values():
            for node in group.nodes:
                self._trace_decorator(node, group)

        return self.clean_groups(result)



from __future__ import annotations

from typing import Any, TypeVar, get_args, get_origin

from dishka.entities.component import Component
from dishka.entities.key import DependencyKey
from dishka.entities.scope import BaseScope
from .factory import Factory
from .type_match import get_typevar_replacement, is_broader_or_same_type


class Decorator:
    __slots__ = ("factory", "provides")

    def __init__(
            self,
            factory: Factory,
            provides: DependencyKey | None = None,
    ) -> None:
        self.factory = factory
        if provides:
            self.provides = provides
        else:
            self.provides = factory.provides

    def is_generic(self) -> bool:
        return (
            isinstance(self.provides.type_hint, TypeVar)
            or get_origin(self.provides.type_hint) is not None
        )

    def match_type(self, type_: Any) -> bool:
        return is_broader_or_same_type(self.provides.type_hint, type_)

    def as_factory(
            self, *,
            scope: BaseScope,
            new_dependency: DependencyKey,
            cache: bool,
            component: Component,
    ) -> Factory:
        typevar_replacement = get_typevar_replacement(
            self.provides.type_hint,
            new_dependency.type_hint,
        )
        return Factory(
            scope=scope,
            source=self.factory.source,
            provides=self.factory.provides.with_component(component),
            is_to_bind=self.factory.is_to_bind,
            dependencies=[
                self._replace_dep(
                    dep, new_dependency, typevar_replacement,
                ).with_component(component)
                for dep in self.factory.dependencies
            ],
            kw_dependencies={
                name: self._replace_dep(
                    dep, new_dependency, typevar_replacement,
                ).with_component(component)
                for name, dep in self.factory.kw_dependencies.items()
            },
            type_=self.factory.type,
            cache=cache,
            override=False,
        )

    def _replace_dep(
            self,
            old_key: DependencyKey,
            new_key: DependencyKey,
            typevar_replacement: dict[TypeVar, Any],
    ) -> DependencyKey:
        if old_key == self.provides:
            return new_key
        if get_origin(old_key.type_hint):
            args = tuple(
                typevar_replacement.get(t, t)
                for t in get_args(old_key.type_hint)
            )
            return DependencyKey(
                old_key.type_hint[args],
                self.provides.component,
            )
        return old_key

    def __get__(self, instance: Any, owner: Any) -> Decorator:
        return Decorator(self.factory.__get__(instance, owner))



from collections import defaultdict
from collections.abc import Sequence
from typing import Any, TypeVar, cast, get_origin

from ._adaptix.type_tools.basic_utils import is_generic
from .dependency_source import (
    Alias,
    ContextVariable,
    Decorator,
    Factory,
)
from .entities.component import DEFAULT_COMPONENT, Component
from .entities.factory_type import FactoryType
from .entities.key import DependencyKey
from .entities.scope import BaseScope, InvalidScopes, Scope
from .entities.validation_settigs import ValidationSettings
from .exceptions import (
    AliasedFactoryNotFoundError,
    CycleDependenciesError,
    GraphMissingFactoryError,
    ImplicitOverrideDetectedError,
    InvalidGraphError,
    NoFactoryError,
    NothingOverriddenError,
    UnknownScopeError,
)
from .provider import BaseProvider
from .registry import Registry

DECORATED_COMPONENT_PREFIX = "__Dishka_decorate_"


class GraphValidator:
    def __init__(self, registries: Sequence[Registry]) -> None:
        self.registries = registries
        self.path: dict[DependencyKey, Factory] = {}
        self.valid_keys: dict[DependencyKey, bool] = {}

    def _validate_key(
            self, key: DependencyKey, registry_index: int,
    ) -> None:
        if key in self.valid_keys:
            return
        if key in self.path:
            keys = list(self.path)
            factories = list(self.path.values())[keys.index(key):]
            raise CycleDependenciesError(factories)
        for index in range(registry_index + 1):
            registry = self.registries[index]
            factory = registry.get_factory(key)
            if factory:
                self._validate_factory(factory, registry_index)
                return
        raise NoFactoryError(requested=key)

    def _validate_factory(
            self, factory: Factory, registry_index: int,
    ) -> None:
        self.path[factory.provides] = factory
        if (
            factory.provides in factory.kw_dependencies.values() or
            factory.provides in factory.dependencies
        ):
            raise CycleDependenciesError([factory])
        try:
            for dep in factory.dependencies:
                # ignore TypeVar parameters
                if not isinstance(dep.type_hint, TypeVar):
                    self._validate_key(dep, registry_index)
            for dep in factory.kw_dependencies.values():
                # ignore TypeVar parameters
                if not isinstance(dep.type_hint, TypeVar):
                    self._validate_key(dep, registry_index)

        except NoFactoryError as e:
            e.add_path(factory)
            raise
        finally:
            self.path.pop(factory.provides)
        self.valid_keys[factory.provides] = True

    def validate(self) -> None:
        for registry_index, registry in enumerate(self.registries):
            factories = tuple(registry.factories.values())
            for factory in factories:
                self.path = {}
                try:
                    self._validate_factory(factory, registry_index)
                except NoFactoryError as e:
                    raise GraphMissingFactoryError(
                        e.requested, e.path,
                        self._find_other_scope(e.requested),
                        self._find_other_component(e.requested),
                    ) from None
                except CycleDependenciesError as e:
                    raise e from None

    def _find_other_scope(self, key: DependencyKey) -> list[Factory]:
        found = []
        for registry in self.registries:
            for factory_key, factory in registry.factories.items():
                if factory_key == key:
                    found.append(factory)
        return found

    def _find_other_component(self, key: DependencyKey) -> list[Factory]:
        found = []
        for registry in self.registries:
            for factory_key, factory in registry.factories.items():
                if factory_key.type_hint != key.type_hint:
                    continue
                if factory_key.component == key.component:
                    continue
                found.append(factory)
        return found


class RegistryBuilder:
    def __init__(
            self,
            *,
            scopes: type[BaseScope],
            providers: Sequence[BaseProvider],
            container_key: DependencyKey,
            skip_validation: bool,
            validation_settings: ValidationSettings,
    ) -> None:
        self.scopes = scopes
        self.providers = providers
        self.registries: dict[BaseScope, Registry] = {}
        self.dependency_scopes: dict[DependencyKey, BaseScope] = {}
        self.components: set[Component] = {DEFAULT_COMPONENT}
        self.alias_sources: dict[DependencyKey, Any] = {}
        self.aliases: dict[DependencyKey, Alias] = {}
        self.container_key = container_key
        self.decorator_depth: dict[DependencyKey, int] = defaultdict(int)
        self.skip_validation = skip_validation
        self.validation_settings = validation_settings
        self.processed_factories: dict[DependencyKey, Factory] = {}

    def _collect_components(self) -> None:
        for provider in self.providers:
            self.components.add(provider.component)

    def _collect_provided_scopes(self) -> None:
        for provider in self.providers:
            for factory in provider.factories:
                if not isinstance(factory.scope, self.scopes):
                    raise UnknownScopeError(factory.scope, self.scopes)
                provides = factory.provides.with_component(provider.component)
                self.dependency_scopes[provides] = factory.scope
            for context_var in provider.context_vars:
                for component in self.components:
                    provides = context_var.provides.with_component(component)
                    # typing.cast is applied because the scope
                    # was checked above
                    self.dependency_scopes[provides] = cast(
                        BaseScope, context_var.scope,
                    )

    def _collect_aliases(self) -> None:
        for provider in self.providers:
            component = provider.component
            for alias in provider.aliases:
                provides = alias.provides.with_component(component)
                alias_source = alias.source.with_component(component)
                self.alias_sources[provides] = alias_source
                self.aliases[provides] = alias

    def _init_registries(self) -> None:
        has_fallback = True
        for scope in self.scopes:
            registry = Registry(scope, has_fallback=has_fallback)
            context_var = ContextVariable(
                provides=self.container_key,
                scope=scope,
                override=False,
            )
            for component in self.components:
                registry.add_factory(context_var.as_factory(component))
            self.registries[scope] = registry
            has_fallback = False

    def _process_factory(
            self, provider: BaseProvider, factory: Factory,
    ) -> None:
        factory = factory.with_component(provider.component)
        provides = factory.provides
        if (
            self.validation_settings.nothing_overridden
            and not self.skip_validation
            and factory.override
            and provides not in self.processed_factories
        ):
            raise NothingOverriddenError(factory)

        if (
            self.validation_settings.implicit_override
            and not self.skip_validation
            and not factory.override
            and provides in self.processed_factories
        ):
            raise ImplicitOverrideDetectedError(
                factory,
                self.processed_factories[provides],
            )

        self.processed_factories[provides] = factory
        registry = self.registries[cast(Scope, factory.scope)]
        registry.add_factory(factory)

    def _process_alias(
            self, provider: BaseProvider, alias: Alias,
    ) -> None:
        component = provider.component
        alias_source = alias.source.with_component(component)
        visited_keys: list[DependencyKey] = []
        while alias_source not in self.dependency_scopes:
            if alias_source not in self.alias_sources:
                if self.skip_validation:
                    return
                e = NoFactoryError(alias_source)
                for key in visited_keys[::-1]:
                    e.add_path(self.aliases[key].as_factory(
                        InvalidScopes.UNKNOWN_SCOPE, key.component,
                    ))
                e.add_path(alias.as_factory(
                    InvalidScopes.UNKNOWN_SCOPE, component,
                ))
                raise e
            visited_keys.append(alias_source)
            alias_source = self.alias_sources[alias_source]
            if alias_source in visited_keys:
                if self.skip_validation:
                    return
                raise CycleDependenciesError([
                    self.aliases[s].as_factory(
                        InvalidScopes.UNKNOWN_SCOPE, component,
                    )
                    for s in visited_keys
                ])

        scope = self.dependency_scopes[alias_source]
        registry = self.registries[scope]

        factory = alias.as_factory(scope, component)
        if (
            self.validation_settings.nothing_overridden
            and not self.skip_validation
            and factory.override
            and factory.provides not in self.processed_factories
        ):
            raise NothingOverriddenError(factory)

        if (
            self.validation_settings.implicit_override
            and not self.skip_validation
            and not factory.override
            and factory.provides in self.processed_factories
        ):
            raise ImplicitOverrideDetectedError(
                factory,
                self.processed_factories[factory.provides],
            )

        self.dependency_scopes[factory.provides] = scope
        self.processed_factories[factory.provides] = factory
        registry.add_factory(factory)

    def _process_generic_decorator(
            self, provider: BaseProvider, decorator: Decorator,
    ) -> None:
        found = []
        provides = decorator.provides.with_component(provider.component)
        for registry in self.registries.values():
            for factory in registry.factories.values():
                if factory.provides.component != provides.component:
                    continue
                if factory.type is FactoryType.CONTEXT:
                    continue
                if decorator.match_type(factory.provides.type_hint):
                    found.append((registry, factory))
        if found:
            for registry, factory in found:
                self._decorate_factory(
                    decorator=decorator,
                    registry=registry,
                    old_factory=factory,
                )
        else:
            if not self.validation_settings.nothing_decorated:
                return
            if self.skip_validation:
                return
            raise GraphMissingFactoryError(
                requested=provides,
                path=[decorator.as_factory(
                    scope=InvalidScopes.UNKNOWN_SCOPE,
                    new_dependency=provides,
                    cache=False,
                    component=provider.component,
                )],
            )

    def _process_normal_decorator(
            self, provider: BaseProvider, decorator: Decorator,
    ) -> None:
        provides = decorator.provides.with_component(provider.component)
        if provides not in self.dependency_scopes:
            if not self.validation_settings.nothing_decorated:
                return
            if self.skip_validation:
                return
            raise GraphMissingFactoryError(
                requested=provides,
                path=[decorator.as_factory(
                    scope=InvalidScopes.UNKNOWN_SCOPE,
                    new_dependency=provides,
                    cache=False,
                    component=provider.component,
                )],
            )
        scope = self.dependency_scopes[provides]
        registry = self.registries[scope]
        # factory is expected to be as we already processed
        # it according to dependency_scopes
        old_factory = cast(Factory, registry.get_factory(provides))
        self._decorate_factory(
            decorator=decorator,
            registry=registry,
            old_factory=old_factory,
        )

    def _is_alias_decorated(
        self,
        decorator: Decorator,
        registry: Registry,
        alias: Factory,
    ) -> bool:
        dependency = alias.dependencies[0]
        factory = registry.get_factory(dependency)
        if factory is None:
            raise AliasedFactoryNotFoundError(dependency, alias)
        return factory.source is decorator.factory.source

    def _decorate_factory(
        self,
        decorator: Decorator,
        registry: Registry,
        old_factory: Factory,
    ) -> None:
        provides = old_factory.provides
        if provides.component is None:
            raise ValueError(  # noqa: TRY003
                f"Unexpected empty component for {provides}",
            )
        if (
            old_factory.type is FactoryType.ALIAS
                and self._is_alias_decorated(decorator, registry, old_factory)
        ):
            return
        depth = self.decorator_depth[provides]
        decorated_component = (f"{DECORATED_COMPONENT_PREFIX}{depth}_"
                               f"{provides.component}")
        self.decorator_depth[provides] += 1
        if old_factory is None:
            raise InvalidGraphError(  # noqa: TRY003
                "Cannot apply decorator because there is"
                f"no factory for {provides}",
            )
        if old_factory.type is FactoryType.CONTEXT:
            raise InvalidGraphError(  # noqa: TRY003
                f"Cannot apply decorator to context data {provides}",
            )
        old_factory.provides = DependencyKey(
            old_factory.provides.type_hint, decorated_component,
        )
        new_factory = decorator.as_factory(
            scope=registry.scope,
            new_dependency=old_factory.provides,
            cache=old_factory.cache,
            component=provides.component,
        )
        new_factory.provides = provides
        registry.add_factory(old_factory)
        registry.add_factory(new_factory)

    def _process_context_var(
            self, provider: BaseProvider, context_var: ContextVariable,
    ) -> None:
        if context_var.scope is None:
            raise UnknownScopeError(
                context_var.scope,
                self.scopes,
                extend_message=(
                    "Define it explicitly in Provider or from_context"
                ),
            )
        registry = self.registries[context_var.scope]
        for component in self.components:
            factory = context_var.as_factory(component)
            if (
                self.validation_settings.nothing_overridden
                and not self.skip_validation
                and factory.override
                and factory.provides not in self.processed_factories
            ):
                raise NothingOverriddenError(factory)

            if (
                self.validation_settings.implicit_override
                and not self.skip_validation
                and not factory.override
                and factory.provides in self.processed_factories
            ):
                raise ImplicitOverrideDetectedError(
                    factory,
                    self.processed_factories[factory.provides],
                )
            self.processed_factories[context_var.provides] = factory
            registry.add_factory(factory)

    def build(self) -> tuple[Registry, ...]:
        self._collect_components()
        self._collect_provided_scopes()
        self._collect_aliases()
        self._init_registries()

        for provider in self.providers:
            for factory in provider.factories:
                self._process_factory(provider, factory)
            for alias in provider.aliases:
                self._process_alias(provider, alias)
            for context_var in provider.context_vars:
                self._process_context_var(provider, context_var)
            for decorator in provider.decorators:
                if decorator.is_generic():
                    self._process_generic_decorator(provider, decorator)
                else:
                    self._process_normal_decorator(provider, decorator)
        self._post_process_generic_factories()
        registries = list(self.registries.values())
        if not self.skip_validation:
            GraphValidator(registries).validate()
        return tuple(registries)

    def _post_process_generic_factories(self) -> None:
        found = [
            (registry, factory)
            for registry in self.registries.values()
            for factory in registry.factories.values()
            if is_generic(factory.provides.type_hint)
        ]
        for registry, factory in found:
            origin = get_origin(factory.provides.type_hint)
            origin_key = DependencyKey(origin, factory.provides.component)
            registry.add_factory(factory, origin_key)



__all__ = [
    "FromDishka",
    "LitestarProvider",
    "inject",
    "inject_websocket",
    "setup_dishka",
]

from collections.abc import Callable
from inspect import Parameter
from typing import ParamSpec, TypeVar, get_type_hints

from litestar import Litestar, Request, WebSocket
from litestar.enums import ScopeType
from litestar.types import ASGIApp, Receive, Scope, Send

from dishka import AsyncContainer, FromDishka, Provider, from_context
from dishka import Scope as DIScope
from dishka.integrations.base import wrap_injection

P = ParamSpec("P")
T = TypeVar("T")


def inject(func: Callable[P, T]):
    return _inject_wrapper(func, "request", Request)


def inject_websocket(func: Callable[P, T]):
    return _inject_wrapper(func, "socket", WebSocket)


def _inject_wrapper(
        func: Callable[P, T],
        param_name: str,
        param_annotation: type[Request | WebSocket],
):
    hints = get_type_hints(func)

    request_param = next(
        (name for name in hints if name == param_name),
        None,
    )

    if request_param:
        additional_params = []
    else:
        additional_params = [Parameter(
            name=param_name,
            annotation=param_annotation,
            kind=Parameter.KEYWORD_ONLY,
        )]

    return wrap_injection(
        func=func,
        is_async=True,
        additional_params=additional_params,
        container_getter=lambda _, r: r[param_name].state.dishka_container,
    )


class LitestarProvider(Provider):
    request = from_context(Request, scope=DIScope.REQUEST)
    socket = from_context(WebSocket, scope=DIScope.SESSION)


def make_add_request_container_middleware(app: ASGIApp) -> ASGIApp:
    async def middleware(scope: Scope, receive: Receive, send: Send) -> None:
        if scope.get("type") not in (ScopeType.HTTP, ScopeType.WEBSOCKET):
            await app(scope, receive, send)
            return

        if scope.get("type") == ScopeType.HTTP:
            request = Request(scope)  # type: ignore[var-annotated]
            context = {Request: request}
            di_scope = DIScope.REQUEST

        else:
            request = WebSocket(scope)
            context = {WebSocket: request}
            di_scope = DIScope.SESSION

        async with request.app.state.dishka_container(
            context, scope=di_scope,
        ) as request_container:
            request.state.dishka_container = request_container
            await app(scope, receive, send)

    return middleware


def setup_dishka(container: AsyncContainer, app: Litestar) -> None:
    app.asgi_handler = make_add_request_container_middleware(
        app.asgi_handler,
    )
    app.state.dishka_container = container



import sys
from typing import Any

from dishka._adaptix.common import TypeHint

if sys.version_info >= (3, 12):
    from typing import TypeAliasType


    def is_type_alias_type(tp: TypeHint) -> bool:
        return isinstance(tp, TypeAliasType)

else:
    def is_type_alias_type(tp: TypeHint) -> bool:
        return False


def unwrap_type_alias(hint: Any) -> Any:
    while is_type_alias_type(hint):
        hint = hint.__value__
    return hint
